/* Finding unordered gene clusters.  One gene may belong to multiple homolgous groups.
 * Author: Qingwu Yang
 * Texas A&M University
 * July 6, 2008
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include <math.h>

#define EXIT0       fflush(NULL), exit(0)
#define FPFE1(x)    fprintf(stdout, "%s", x), fflush(NULL), exit(0)
#define FPFE2(x, y) fprintf(stdout, x, y), fflush(NULL), exit(0)

int  MIN_SET_OUTPUT = 1;     //the minimal size of subsets that will be output
double EVALUE_CUTOFF = 1e-1;  //cutoff for e-values
int  MIN_N = 2;              //the minimal number of chromosomes
int  IF_CIRCULAR_GENOME = 0; 
int     ave_L;

int   MAX_HOMOLOGIDS = 40;
int   *ChrNO; //current chr# of each genome

typedef struct ssetindex {
	int  a;
	int  n;
} ssetindex;

typedef struct rnode {
	int   *set;
	double eval;      //evalues  = EVALUE[avl][n]
	int    l;         //# of unique clusters
	int    n;         //# of chromosomes that contain this cluster
	int    avl;       //average length of start-end distances
//	int    avcl;      //average length of chromosomes
	char ** s;
	int    *chr;      //chromosome ID's
	int    *low;      //starting point on the first chromosome (reference window)
	int    *high;     //end point on the first chromosome
	int    *numg;     //# of genes on each chromosome
	int   **posg;     //gene positions on each chromosome
	struct rnode  *next;
} rnode;


typedef struct bigdouble {
	double d;
	double e;
} bigdouble;


rnode *rnhead, *rntail, *rnstart, *rnend, *rncurr, *rnnew;

//info()
void info();

//parse commandline
void cmd_parser(int ac, char ** av, int * k, char * fname, char *ofname);

//check if a line is blank
int is_blank(char * s);

//get file names containing input data
void get_file_names(char * fname, char *** genomenames, int *num_genomes);

//read in data
void data_reader(int ****gchrs, int **num_genes, char **genomenames, int num_genomes);

//data transform --- gene -- poss
void data_transform(int **** xposs, int *gmax, int n_genomes, int * n_genes, int *** xchrs);

//search for syntenies
void syn_finder(int *** xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene, int xmax, int k, char *ofname);

//search for syntenies -- for circular genomes
void syn_finder_circular(int *** xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene, int xmax, int k, char *ofname);

//find intersections using brute force approach
void brute_force_insection(int ***xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene,
	                       int xmax, int k, char *ofname, int *bigmem, ssetindex **subs, int *n_subs);

int ** tmp_K;
int *  n_tmp_K;

int  * cstat; //array for fast intersection

double ** EVALUE;
double ***allEVALUES;

//calculate avervage length
int ave_l(int * n_genes, int n_genomes);

//calculate e-values
void calcultate_EVALUE(int k, int l, int n);

//calculate all e-values
void calcultate_all_EVALUE(int l, int n);

//calculate all e-values -- fast
void calcultate_ALL_EVALUE_FAST(int l, int n);

//calculate all e-values
void comput_ALL_EVALUE(int l, int n, int r);

//reorganize the output results
void reorganize_output(int ng, char *tfname, char * ofname);

//reorganize the output results
void reorganize_output_2(int ng, char *tfname);

int main(int argc, char ** argv){
	int  num_genomes, *num_genes, gmax;
	int  *** gchrs;       //genes on all chromosomes [#chr][#gene][#homolog]. 
						  //Note. if a gene doesn't have homologs at all, [#chr][#gene]=null
	                      //Otherwise, [#chr][#gene][0] is number of homolog ids
	int  *** gposs;       //gene positions on the genomes
	char ** gnames;      //gene names
	char ** genomenames; //genome names
	int     dist_k;      //distance cutoff
	char    datafname[30]; //name of file containing input data
	char    outfname[30];  //name of file containing output
	char    tmpoutfname[40];

	int i; //,k;,j
	time_t  time1, time2;

	//parse commandline
	strcpy(outfname, "syn_result");
	strcpy(datafname, "filelist.txt");
	cmd_parser(argc, argv, &dist_k, datafname, outfname);
	strcpy(tmpoutfname, outfname);
	strcat(tmpoutfname, "_tmpxyz");

	//get file names containing input data
	get_file_names(datafname, &genomenames, &num_genomes);

	//read in data -- genes on chromosomes in order
	data_reader(&gchrs, &num_genes, genomenames, num_genomes);

	/*
	{
		FILE * test;
		int j;

		test = fopen("test", "w");
		for(i=0; i<=num_genes[2]; ++i){
			if( gchrs[2][i] ){
				fprintf(test, "\n%5d ", i);

				for(j=1; j<=gchrs[2][i][0]; ++j)
					fprintf(test, " %5d", gchrs[2][i][j]);
			}
		}
		fclose(test);
	}
	//*/	

	//calcualte e-values
	ave_L = ave_l(num_genes, num_genomes);
	//calcultate_EVALUE(dist_k+1, ave_L, num_genomes);
	time1 = time(NULL);
	//calcultate_ALL_EVALUE_FAST(ave_L, num_genomes);
	comput_ALL_EVALUE(ave_L, num_genomes, dist_k+1);
	time2 = time(NULL);
	//fprintf(stdout, "\ncalculate_all_EVALUE: %7.2f mins\n", (double)difftime(time2, time1)/60), fflush(NULL);

	//data transform --- gene -- poss
	data_transform(&gposs, &gmax, num_genomes, num_genes, gchrs);

	//search for syntenies
	gnames = NULL;
	remove(tmpoutfname);

	i = dist_k;
	//printf("\ni=%d  ( %d )", i, dist_k), fflush(NULL);
	EVALUE = allEVALUES[i];

	if(  IF_CIRCULAR_GENOME == 1 ){ //search for syntenies -- for circular genomes
		syn_finder_circular(gchrs, gposs, gnames, genomenames, num_genomes, num_genes, gmax, dist_k, tmpoutfname);
	} else 
		syn_finder(gchrs, gposs, gnames, genomenames, num_genomes, num_genes, gmax, dist_k, tmpoutfname);

	//remove repeats and subsets
	reorganize_output(num_genomes, tmpoutfname, outfname);

	remove(tmpoutfname);

	return 0;
} //main()

//calculate avervage length
int ave_l(int * n_genes, int n_genomes){
	int i,j;

	for(i=0, j=0; i<n_genomes; i++)
		j += n_genes[i];

	return (j/n_genomes);
}

//search for syntenies
void syn_finder(int *** xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene, int xmax, int k, char *ofname){
	int b,d, i, j, p, q,x, y,z, cmax,m, K, low, high;
	int *bigmem, memsize;  //big memory for intermediate subsets

	ssetindex ** subs;     //positions of all subsets
	int       *  n_subs;   //# of subsets
	int       *  currk;
	int          n_currk;
	int  yqi;

	K = k+1;
	//get the maximal number of paralogs
	cmax = 0;
	for(i=0; i<=xmax; i++){
		for(j=0; j<n_genome; j++)
			if( xpos[i][j] != NULL && xpos[i][j][0] >cmax )
				cmax = xpos[i][j][0];
	}

	//allocate space for bigmem
	memsize = n_genome*K*K*cmax/2;

	if( !(bigmem = calloc(memsize, sizeof(int))) )
		fprintf(stdout, "\nmemsize=%d \nout of memory for bigmem 0 \n", memsize), fflush(NULL), exit(0);
	//allocate space for subs and n_subs
	if( !(subs = malloc(n_genome * sizeof(ssetindex *))) )
		FPFE1("out of memory for subs \n");
	m = K*cmax*MAX_HOMOLOGIDS;

	for(i=0; i<n_genome; i++){
		if( !(subs[i] = malloc(m * sizeof(ssetindex))) )
			FPFE1("out of memory for subs[]\n");
	}
	if( !(n_subs = malloc(n_genome*sizeof(int))) )
		FPFE1("out of memory for n_subs\n");

	//allocate space for currk
	if( !(currk = malloc(m * sizeof(int))) )
		FPFE1("out of memory for currk\n");

	//allocate space for global tmp_K
	if( !(tmp_K = malloc( (n_genome+1) * sizeof(int *))) )
		FPFE1("out of memory for tmp_K\n");
	for(i=0; i<=n_genome; i++){
		if( !(tmp_K[i] = malloc(K*MAX_HOMOLOGIDS * sizeof(int))) )
			FPFE1("out of memory for tmp_K[]\n");
	}
	if( !(n_tmp_K = malloc( (n_genome+1) * sizeof(int))) )
		FPFE1("out of memory for n_tmp_K\n");

	if( !(cstat = malloc( (xmax+1) *sizeof(int))) )
		FPFE1("out of memory for cstat\n");

	//main loop
	for(i=0; i<=n_genome-MIN_N; i++){ //start search from each genome once
	//for(i=0; i<1; i++){ //start search from each genome once
		for(j=0; j<n_genome; j++)
			n_subs[j] = 0;

		low=high=0;
		for(j=1; j<=n_gene[i]; j++){ //scan a genome position by position
			while( j<=n_gene[i] && !xchr[i][j] ) 
				j++;
			if ( j > n_gene[i] )
				break;

			n_subs[i] = 1;
			q=0;
			subs[i][0].a = 0;
			subs[i][0].n = 1;
			bigmem[q]    = j;
			q++;

			low = j;

			for(p=1; p<K && p+j<=n_gene[i]; p++){
				if( xchr[i][j+p] ){
					low = j+p;
					bigmem[q] = low;
					(subs[i][0].n) ++;
					q++;
				}
			}

			if( low > high) {//new doable k-set
				high = low;

				for(x=i+1; x<n_genome; x++){//find all subsets
					n_subs[x] = 0;

					//put locations of all genes in the current k-set into currk
					n_currk=0;
					for(y=0; y<subs[i][0].n; y++){
						for(yqi=1; yqi<= xchr[i][bigmem[y]][0]; ++ yqi){
							if( xpos[ xchr[i][bigmem[y]][yqi] ][x] != NULL ){
								for(z=1; z<=xpos[ xchr[i][bigmem[y]][yqi] ][x][0]; z++){
									currk[n_currk] = xpos[ xchr[i][bigmem[y]][yqi] ][x][z];
									n_currk ++;
								}
							}
						}						
					}

					if( n_currk > 0 ){ //not empty
						//sort currk
						for(y=0; y<n_currk; y++){
							b=currk[y];
							for(z=y+1; z<n_currk; z++){
								if(currk[z] < b){
									b = currk[z];
									d = z;
								}
							}
							if( b < currk[y] ){
								currk[d] = currk[y];
								currk[y] = b;
							}
						}

						//remove repeats in currk
						for(y=1; y<n_currk; y++){
							if( currk[y] == currk[y-1] ){
								for(z=y; z<n_currk-1; z++)
									currk[z] = currk[z+1];
								n_currk--;
								y--;
							}
						}

						//find k-sets
						d=-1;
						for(y=0; y<n_currk; y++){ //scan currk position by position
							z=y+1;
							while( z<n_currk && currk[z]-currk[y]<=k )
								z++;
							if( z-1 > d ){
								d = z-1;
								(subs[x][n_subs[x]]).a = q;
								(subs[x][n_subs[x]]).n = 0;
								for(b=y; b<z; b++){
									if( q == memsize ){  //bigmem is full. add space
										FPFE1("Go out of memory for bigmem! \n");
									}
									//bigmem[q] = xchr[x][currk[b]];
									bigmem[q] = currk[b];
									q++;
									((subs[x][n_subs[x]]).n)++;
								}
								n_subs[x] ++;
							}
						} //scan currk position by position
					}////if( n_currk > 0 ){ //not empty
				}////for(x=i+1; x<n_genome; x++){//find all subsets

				//try to find all intersections and output the results
				brute_force_insection(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs);

			}////if( low > high) {//new doable k-set
		}////for(j=1; j<=n_gene[i]; j++){ //scan a genome position by position
	}////start search from each genome once

	if( bigmem )
		free( bigmem );
	for(i=0; i<n_genome; i++){
		if( subs[i] )
			free( subs[i] );
	}
	if( subs )
		free( subs );
	if( n_subs )
		free( n_subs );
	if( currk )
		free( currk );
	for(i=0; i<=n_genome; i++){
		if( tmp_K[i] )
			free( tmp_K[i] );
	}
	if( tmp_K )
		free( tmp_K );
	if( n_tmp_K )
		free( n_tmp_K );

	if( cstat )
		free( cstat );

	return;
}

//>>>OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO
//search for syntenies -- for circular genomes  -- Jun-25-06
void syn_finder_circular(int *** xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene, int xmax, int k, char *ofname){
	int b,d, i, j, p, q,x, y,z, cmax,m, K, low, high;
	int *bigmem, memsize;  //big memory for intermediate subsets

	ssetindex ** subs;     //positions of all subsets
	int       *  n_subs;   //# of subsets
	int       *  currk;
	int          n_currk;
	int rest, rest2;
	int yqi;

	K = k+1;
	//get the maximal number of paralogs
	cmax = 0;
	for(i=0; i<=xmax; i++){
		for(j=0; j<n_genome; j++)
			if( xpos[i][j] != NULL && xpos[i][j][0] >cmax )
				cmax = xpos[i][j][0];
	}

	//allocate space for bigmem
	memsize = n_genome*K*K*cmax/2;

	if( !(bigmem = calloc(memsize, sizeof(int))) )
		fprintf(stdout, "\nmemsize=%d \nout of memory for bigmem 0 \n", memsize), fflush(NULL), exit(0);
	//allocate space for subs and n_subs
	if( !(subs = malloc(n_genome * sizeof(ssetindex *))) )
		FPFE1("out of memory for subs \n");
	m = K*cmax*MAX_HOMOLOGIDS;

	for(i=0; i<n_genome; i++){
		if( !(subs[i] = malloc(m * sizeof(ssetindex))) )
			FPFE1("out of memory for subs[]\n");
	}
	if( !(n_subs = malloc(n_genome*sizeof(int))) )
		FPFE1("out of memory for n_subs\n");

	//allocate space for currk
	if( !(currk = malloc(m * sizeof(int))) )
		FPFE1("out of memory for currk\n");

	//allocate space for global tmp_K
	if( !(tmp_K = malloc( (n_genome+1) * sizeof(int *))) )
		FPFE1("out of memory for tmp_K\n");
	for(i=0; i<=n_genome; i++){
		if( !(tmp_K[i] = malloc(K*MAX_HOMOLOGIDS * sizeof(int))) )
			FPFE1("out of memory for tmp_K[]\n");
	}
	if( !(n_tmp_K = malloc( (n_genome+1) * sizeof(int))) )
		FPFE1("out of memory for n_tmp_K\n");

	if( !(cstat = malloc( (xmax+1) *sizeof(int))) )
		FPFE1("out of memory for cstat\n");

	//main loop
	for(i=0; i<=n_genome-MIN_N; i++){ //start search from each genome once
	//for(i=0; i<1; i++){ //start search from each genome once
		for(j=0; j<n_genome; j++)
			n_subs[j] = 0;

		low=high=0;
		for(j=1; j<=n_gene[i]; j++){ //scan a genome position by position
			while( j<=n_gene[i] && !xchr[i][j] ) 
				j++;
			if ( j > n_gene[i] )
				break;
			
			n_subs[i] = 1;
			q=0;
			subs[i][0].a = 0;
			subs[i][0].n = 1;
			bigmem[q]    = j;
			q++;

			low = j;

			//for(p=1; p<K && p+j<=n_gene[i]; p++){
			for(p=1; p<K; p++){
				rest2 = j+p;
				rest = (rest2 > n_gene[i]) ? (rest2 % n_gene[i]) : rest2;

				if( xchr[i][rest] ){
					low = rest2;
					bigmem[q] = rest;
					(subs[i][0].n) ++;
					q++;
				}
			}

			if( low > high) {//new doable k-set
				high = low;

				for(x=i+1; x<n_genome; x++){//find all subsets
					n_subs[x] = 0;

					//put locations of all genes in the current k-set into currk
					n_currk=0;
					for(y=0; y<subs[i][0].n; y++){
						for(yqi=1; yqi<= xchr[i][bigmem[y]][0]; ++ yqi){
							if( xpos[ xchr[i][bigmem[y]][yqi] ][x] != NULL ){
								for(z=1; z<=xpos[ xchr[i][bigmem[y]][yqi] ][x][0]; z++){
									currk[n_currk] = xpos[ xchr[i][bigmem[y]][yqi] ][x][z];
									n_currk ++;
								}
							}
						}						
					}

					if( n_currk > 0 ){ //not empty
						//sort currk
						for(y=0; y<n_currk; y++){
							b=currk[y];
							for(z=y+1; z<n_currk; z++){
								if(currk[z] < b){
									b = currk[z];
									d = z;
								}
							}
							if( b < currk[y] ){
								currk[d] = currk[y];
								currk[y] = b;
							}
						}

						//remove repeats in currk
						for(y=1; y<n_currk; y++){
							if( currk[y] == currk[y-1] ){
								for(z=y; z<n_currk-1; z++)
									currk[z] = currk[z+1];
								n_currk--;
								y--;
							}
						}

						//find k-sets
						d=-1;
						for(y=0; y<n_currk; y++){ //scan currk position by position
							z = y+1;

							while(((z >= n_currk)?(n_gene[x] + currk[z % n_currk]) : (currk[z])) - currk[y]<=k )
								++ z;
							
							//z=y+1;
							//while( z<n_currk && currk[z]-currk[y]<=k ) z++;
							if( z-1 > d ){
								d = z-1;
								(subs[x][n_subs[x]]).a = q;
								(subs[x][n_subs[x]]).n = 0;
								for(b=y; b<z; b++){
									if( q == memsize ){  //bigmem is full. add space
										FPFE1("Go out of memory for bigmem! \n");
									}

									bigmem[q] = (b<n_currk) ? (currk[b]) : (currk[ b%n_currk ]);									
									//bigmem[q] = currk[b];
									q++;
									((subs[x][n_subs[x]]).n)++;
								}
								n_subs[x] ++;
							}
						} //scan currk position by position
					}////if( n_currk > 0 ){ //not empty
				}////for(x=i+1; x<n_genome; x++){//find all subsets

				//try to find all intersections and output the results
				brute_force_insection(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs);

			}////if( low > high) {//new doable k-set
		}////for(j=1; j<=n_gene[i]; j++){ //scan a genome position by position
	}////start search from each genome once

	if( bigmem )
		free( bigmem );
	for(i=0; i<n_genome; i++){
		if( subs[i] )
			free( subs[i] );
	}
	if( subs )
		free( subs );
	if( n_subs )
		free( n_subs );
	if( currk )
		free( currk );
	for(i=0; i<=n_genome; i++){
		if( tmp_K[i] )
			free( tmp_K[i] );
	}
	if( tmp_K )
		free( tmp_K );
	if( n_tmp_K )
		free( n_tmp_K );

	if( cstat )
		free( cstat );

	return;
}



//find intersections using a fast (O(nk)) approach  -- Jun-25-06
void bf_intersect_fast(int *** xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene,
	                       int xmax, int k, char *ofname, int *bigmem, ssetindex **subs, int *n_subs){
	FILE * outf;
	int  zeros, m, n, i, j, x,  count;
	int  kp, np;
	int yqi;

	zeros=0;
	while( zeros<n_genome && n_tmp_K[zeros] == 0 )
		zeros++;

	if( zeros == n_genome )
		return;

	m = n_tmp_K[zeros];
	x = zeros;

	count=0; //number of unique clusterID's
	for(i=0; i<m; i++){
		for( yqi=1; yqi<=xchr[x][tmp_K[x][i]][0]; ++yqi){
			if( cstat[ xchr[x][tmp_K[x][i]][yqi] ] == 0 ){
				cstat[ xchr[x][tmp_K[x][i]][yqi] ] = 1;
			}
		}

		/*
		if( cstat[ xchr[x][tmp_K[x][i]] ] == 0 ){
			cstat[ xchr[x][tmp_K[x][i]] ] = 1;
		}
		*/
	}

	for(i=zeros+1; i<n_genome; i++){
		if( n_tmp_K[i] == 0 ){
			zeros ++;
		} else {
			j = i - zeros;
			for(n=0; n<n_tmp_K[i]; n++){
				for( yqi=1; yqi<=xchr[i][tmp_K[i][n]][0]; ++yqi){
					if( cstat[ xchr[i][tmp_K[i][n]][yqi] ] == j )
						cstat[ xchr[i][tmp_K[i][n]][yqi] ] ++;
				}
			}

			count=0;
			for(n=0; n<m; n++){
				for( yqi=1; yqi<=xchr[x][tmp_K[x][n]][0]; ++yqi){		
					if( cstat[ xchr[x][tmp_K[x][n]][yqi] ] <= j ){
						cstat[ xchr[x][tmp_K[x][n]][yqi] ] = 0;
					} else
						count++;
				}
			}

			if( count == 0 ){ //subset is already empty
				zeros = n_genome;
				break;
			}
		}
	}

	//get the list of unique homolog ids in current intersection
	for(j=n=0; n<m; n++){
		for( yqi=1; yqi<=xchr[x][tmp_K[x][n]][0]; ++yqi){	
			if( cstat[ xchr[x][tmp_K[x][n]][yqi] ] > 0 ){
				for(i=0; i<j; i++){
					if( xchr[x][tmp_K[x][n]][yqi] == tmp_K[n_genome][i] )
						break;
				}
				if( i == j ){
					tmp_K[n_genome][j] = xchr[x][tmp_K[x][n]][yqi];
					j++;
				}
			}
		}
	}
	count=j;

	if( count >= MIN_SET_OUTPUT ){
		int S, E, tt=0;

		////only if the intersection, in which at least one window size is k, will be interesting to be worked on further.
		for(i=0; i<n_genome; i++){
			if(n_tmp_K[i] != 0){				
				S=SHRT_MAX; E=-1;

				for(j=0; j<n_tmp_K[i]; j++){
					for( yqi=1; yqi<=xchr[i][tmp_K[i][j]][0]; ++yqi){
						if( cstat[ xchr[i][tmp_K[i][j]][yqi] ] ){
							if( S > tmp_K[i][j] )
								S = tmp_K[i][j];

							if( E < tmp_K[i][j] )
								E = tmp_K[i][j];
						}
					}
				}

				if( E - S == k ){
					tt = 1;
					break;
				}
			}
		}

		if( ! tt ){
			for(i=0; i<m; i++){
				for( yqi=1; yqi<=xchr[x][tmp_K[x][i]][0]; ++yqi){
					if( cstat[ xchr[x][tmp_K[x][i]][yqi] ] != 0 )
						cstat[ xchr[x][tmp_K[x][i]][yqi] ] = 0;
				}
			}

			return;
		}
		////


		kp = np = 0;
		for(i=0; i<n_genome; i++){
			if(n_tmp_K[i] != 0){
				for(j=0; j<n_tmp_K[i]; j++){
					for( yqi=1; yqi<=xchr[i][tmp_K[i][j]][0]; ++yqi){
						if( cstat[ xchr[i][tmp_K[i][j]][yqi] ] > 0 ){
							kp++;
							break;
						}
					}
				}
				np ++;
			}
		}

		kp = kp/np;

		if( EVALUE[kp][np] <= EVALUE_CUTOFF ){
			int xh, xp;

			if( !(outf = fopen(ofname, "a")) )
				FPFE2("File %s can't be opened for appending\n", ofname);

			fprintf(outf, "\n\nCluster: ");

			//sort the array
			for(i=0; i<count; i++){
				 xh =  tmp_K[n_genome][i];
				 for(j=i+1; j<count; j++){
					 if( tmp_K[n_genome][j] < xh ){
						 xh = tmp_K[n_genome][j];
						 xp = j;
					 }
				 }
				 if( xh < tmp_K[n_genome][i]){
					 tmp_K[n_genome][xp] = tmp_K[n_genome][i];
					 tmp_K[n_genome][i]  = xh;
				 }
			}

			for(i=0; i<count; i++)
				fprintf(outf, " %1d", tmp_K[n_genome][i]);

			fprintf(outf, "\n\tExpect = %10.4e, Size = %1d/%1d, Appear = %1d/%1d", EVALUE[kp][np], count, ave_L, n_genome-zeros, n_genome);
			//*
			for(i=0; i<n_genome; i++){
				if(n_tmp_K[i] != 0){
					S=SHRT_MAX; E=-1;

					for(j=0; j<n_tmp_K[i]; j++){
						for( yqi=1; yqi<=xchr[i][tmp_K[i][j]][0]; ++yqi){
							if( cstat[ xchr[i][tmp_K[i][j]][yqi] ] ){
								if( S > tmp_K[i][j] )
									S = tmp_K[i][j];

								if( E < tmp_K[i][j] )
									E = tmp_K[i][j];
							}
						}
					}
					
					fprintf(outf, "\n\t%15s (Chr:%1d S%1d E%1d):", xnomename[i], ChrNO[i], S, E);
					for(j=0; j<n_tmp_K[i]; j++){
						for( yqi=1; yqi<=xchr[i][tmp_K[i][j]][0]; ++yqi){
							if( cstat[ xchr[i][tmp_K[i][j]][yqi] ] ){
								fprintf(outf, " %2d (%2d)", xchr[i][tmp_K[i][j]][yqi], tmp_K[i][j]);
								//break;
							}
						}					
					}
				} else {
					fprintf(outf, "\n@@");
				}
			}
			//*/
			fclose(outf);
		}//if( EVALUE[kp][np] <= EVALUE_CUTOFF )
	}

	for(i=0; i<m; i++){
		for( yqi=1; yqi<=xchr[x][tmp_K[x][i]][0]; ++yqi){
			if( cstat[ xchr[x][tmp_K[x][i]][yqi] ] != 0 )
				cstat[ xchr[x][tmp_K[x][i]][yqi] ] = 0;
		}
	}

	return;
}


//recurrsive -- Jun-25-06
void bf_recurr(int *** xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene,
	        int xmax, int k, char *ofname, int *bigmem, ssetindex **subs, int *n_subs, int n){
	int i,j, x, y, z;

	if( n_subs[n] != 0 ){
		for(i=0; i<n_subs[n]; i++){
			n_tmp_K[n] = subs[n][i].n;
			for(j=0; j<n_tmp_K[n]; j++)
				tmp_K[n][j] = bigmem[subs[n][i].a + j];

			if( n == n_genome-1 ){
				y=z=0;
				for(x=0; x<n_genome; x++){
					if( n_tmp_K[x] != 0 ){
						y++;
						z += n_tmp_K[x];
					}
				}

				if( y >= MIN_N && EVALUE[z/y][y] <= EVALUE_CUTOFF )
					bf_intersect_fast(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs);
			} else {
				bf_recurr(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs, n+1);
			}
		}

		//dummy case
		n_tmp_K[n] = 0;

		if( n == n_genome-1 ){
			y=z=0;
			for(x=0; x<n_genome; x++){
				if( n_tmp_K[x] != 0 ){
					y++;
					z += n_tmp_K[x];
				}
			}

			if( y >= MIN_N && EVALUE[z/y][y] <= EVALUE_CUTOFF )
				bf_intersect_fast(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs);
		} else {
			bf_recurr(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs, n+1);
		}
	} else {
		n_tmp_K[n] = 0;

		if( n == n_genome-1 ){
			y=z=0;
			for(x=0; x<n_genome; x++){
				if( n_tmp_K[x] != 0 ){
					y++;
					z += n_tmp_K[x];
				}
			}

			if( y >= MIN_N && EVALUE[z/y][y] <= EVALUE_CUTOFF )
				bf_intersect_fast(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs);
		} else {
			bf_recurr(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs, n+1);
		}
	}
}

//find intersections using brute force approach -- Jun-25-06
void brute_force_insection(int *** xchr, int ***xpos, char **xname, char **xnomename, int n_genome, int *n_gene,
	                       int xmax, int k, char *ofname, int *bigmem, ssetindex **subs, int *n_subs){

	int  zeros, i;

	for(i=0; i<=xmax; i++)
		cstat[i] = 0;

	zeros=0;
	while( n_subs[zeros] == 0 )
		zeros++;

	bf_recurr(xchr, xpos, xname, xnomename, n_genome, n_gene, xmax, k, ofname, bigmem, subs, n_subs, zeros);
}


//data transform --- gene -- poss
void data_transform(int **** xposs, int *gmax, int n_genomes, int * n_genes, int *** xchrs){
	int max, i, j, k, m;
	int *** pos;
	int * tmp;

	max = 0;
	for(i=0; i<n_genomes; i++)
		for(j=0; j<=n_genes[i]; j++){
			if( xchrs[i][j] ){
				for(k=1; k<=xchrs[i][j][0]; ++k){
					if( max < xchrs[i][j][k] )
						max = xchrs[i][j][k];
				}
			}
		}
	*gmax = max;
	if( !(pos = malloc((max+1) * sizeof(int **))) )
		FPFE1("out of memory for pos\n");
	for(i=0; i<=max; i++){
		if( !(pos[i] = malloc(n_genomes * sizeof(int *))) )
			FPFE1("out of memory for pos[]\n");
		for(j=0; j<n_genomes; j++)
			pos[i][j] = NULL;
	}
	if( !(tmp = malloc((max+1) * sizeof(int))) )
		FPFE1("out of memory for tmp in data_transform()\n");

	for(i=0; i<n_genomes; i++){
		for(m=0; m<=max; m++)
			tmp[m] = 0;

		for(j=0; j<=n_genes[i]; j++){
			if( xchrs[i][j] ){
				for(k=1; k<=xchrs[i][j][0]; ++k){
					tmp[ xchrs[i][j][k] ] ++;
				}
			}
		}

		for(j=0; j<=max; j++){
			if( tmp[j] > 0 ){
				if( ! (pos[j][i] = malloc( (tmp[j]+1) * sizeof(int))) )
					FPFE1("out of memory for pos[][]\n");
				pos[j][i][0] = tmp[j];
			}
		}

		for(m=0; m<=max; m++)
			tmp[m] = 1;
		for(j=0; j<=n_genes[i]; j++){
			if( xchrs[i][j] ){
				for(k=1; k<=xchrs[i][j][0]; ++k){
					pos[ xchrs[i][j][k] ][i][ tmp[xchrs[i][j][k]] ] = j;
					tmp[ xchrs[i][j][k] ] ++;
				}
			}			
		}
	}

	*xposs = pos;

	if( tmp != NULL )
		free(tmp);

	return;
}


//read in data
void data_reader(int ****gchrs, int **num_genes, char **genomenames, int num_genomes){
	FILE *inf;
	int *** xchrs;
	int *  xnum;
	int i,j, n, m, max, k;
	char buff[4096], buff2[4096];
	char *charptr;

	if( !(xnum = malloc(num_genomes * sizeof(int))) )
		FPFE1("out of memory for xnum\n");
	if( !(xchrs = malloc(num_genomes * sizeof(int **))) )
		FPFE1("out of memory for xchrs\n");

	for(i=0; i<num_genomes; i++){
		if( !(inf = fopen(genomenames[i], "r")) )
			FPFE2("File %1s can't be opened for reading!\n", genomenames[i]);

		max = 0;
		while ( fgets(buff, 4096, inf) ){
			if( !is_blank(buff) ){
				if( sscanf(buff, " %d %d %*s ", &m, &n) != 2 ){
					FPFE2("\nData error in line %1s ", buff);
					FPFE2(" In file %1s\n", genomenames[i]);
				}
				if( max < m )
					max = m;
			}
		}

		xnum[i] = max;
		if( !(xchrs[i] = malloc( (max+1) * sizeof(int *) )) )
			FPFE1("out of memory for xchrs[]\n");

		for(j=0; j<=max; j++)
			xchrs[i][j] = NULL;

		rewind(inf);
		while ( fgets(buff, 4096, inf) ){
			if( is_blank(buff) )
				continue;

			strcpy(buff2, buff);

			charptr = strtok(buff2, " \t\n");
			m = atoi(charptr);
			k=0;
			while( (charptr = strtok(NULL, " \t\n") ) )
				k++;

			if( !(xchrs[i][m] = malloc( (k+1) * sizeof(int) )) )
				FPFE1("out of memory for xchrs[][]\n");
			xchrs[i][m][0] = k;

			strcpy(buff2, buff);
			charptr = strtok(buff2, " \t\n");			
			k=1;
			while( (charptr = strtok(NULL, " \t\n") ) ){
				xchrs[i][m][k] = atoi(charptr);

				k++;
			}
		}

		fclose(inf);
	}

	*gchrs= xchrs;
	*num_genes = xnum;

	return;
}

//get file names containing input data
void get_file_names(char * fname, char *** genomenames, int *num_genomes){
	FILE *inf;
	char buff[4096], *ptr;
	char ** gns;
	int n=0;

	if( !(inf = fopen(fname, "r")) )
		FPFE2("File %1s can't be opened for reading", fname);
	while( fgets(buff, 4096, inf) )
		if( !is_blank(buff) )
			n++;

	*num_genomes = n;
	if( !(gns = malloc(n * sizeof(char *))) )
		FPFE1("out of memory for *genomenames!\n");

	rewind(inf);
	n=0;
	while( fgets(buff, 4096, inf) ){
		if( !is_blank(buff) ){
			ptr=buff;
			while( *ptr != '\0' && *ptr != '\n' && *ptr != EOF )
				ptr++;
			if( *ptr == '\n' || *ptr == EOF )
				*ptr = '\0';

			if( !( gns[n] = malloc( (strlen(buff)+1) * sizeof(char))) )
				FPFE1("out of memory for *genomenames[]!\n");
			strcpy(gns[n], buff);
			n++;
		}
	}
	*genomenames = gns;
	fclose(inf);
	return;
}

//check if a line is blank
int is_blank(char * s){
	while( *s != '\0' && *s != EOF ){
		if( isalnum(*s) )
			return 0;
		s++;
	}
	return 1;
}

//info()
void info(){
	fprintf(stdout, "\nsyn -i=inputfile -f=linear_or_circular_genome -k=distance_cutoff -n=min_set_to_be_output -e=e_vaule_cutoff -o=outputfile -c=chr#\n\n");
	fflush(NULL);
	exit(0);
}

//parse commandline
void cmd_parser(int ac, char ** av, int * k, char * fname, char *ofname){
	if( ac < 3 ){
		info();
	} else {
		ac--;
		while( ac > 0 ){
			if( strstr(av[ac], "-i") ){
				if( sscanf(av[ac], "-i=%s", fname) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-o=") ){
				if( sscanf(av[ac], "-o=%s", ofname) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-f=") ){
				if( sscanf(av[ac], "-f=%d", &IF_CIRCULAR_GENOME) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-k=") ){
				if( sscanf(av[ac], "-k=%d", k) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			}  else if( strstr(av[ac], "-a=") ){
				if( sscanf(av[ac], "-a=%d", &MIN_N) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-n=") ){
				if( sscanf(av[ac], "-n=%d", &MIN_SET_OUTPUT) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-e=") ){
				if( sscanf(av[ac], "-e=%lf", &EVALUE_CUTOFF) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-c=") ){
				int i,j;
				char *p, *q;
				p=av[ac];
				j=0;
				while( strstr(p, "x") ){
					j++;
					p++;
				}

				if( !(ChrNO = malloc(j * sizeof(int))) )
					FPFE1("out of memory for ChrNO\n");
				p=av[ac];
				p=strstr(p, "x");
				q=strtok(p, "x");
				ChrNO[0] = atoi(q);
				i=1;
				while( (q=strtok(NULL, "x")) ){
					ChrNO[i] = atoi(q);
					i++;
				}
			} else {
				fprintf(stdout, "wrong command line input: %s\n", av[ac]);
				info();
			}
			ac--;
		}
	}

	return;
}


//pppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppp
//pppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppp
//calculate combination
void comb(bigdouble ** c, int l){
	bigdouble * b; 
	double x;
	int y, k;

	if( !(b= calloc(l+1, sizeof(bigdouble))) )
		FPFE1("out of memory for b");
	
	b[0].d = 1.0;
	b[0].e = 0;
	for(k=1; k<=l/2; k++){
		x = (double)(l-k+1)/k;

		if( x == 0.0 )
			fprintf(stdout, "\nl=%d k=%d", l, k), fflush(NULL), exit(0);

		y = 0;
		if( x >= 10.0 ){
			while(x >= 10.0){
				x /= 10.0;
				y++;
			}
		} else if ( x < 1.0 ) {
			while( x < 1.0){
				x *= 10.0;
				y--;
			}
		}	
		b[k].e = b[k-1].e + y;
		b[k].d = b[k-1].d * x;

		if( b[k].d >= 10.0 ){
			b[k].d /= 10.0;
			b[k].e += 1;
		}
		//b[k] = (double)(l-k+1)/k * b[k-1];
		//if( l == 2 )
		//	fprintf(stdout, "\nb[%3d]= %2.5e, %2.5e", k, b[k].d, b[k].e);
	}

	for(k=l/2+1; k<=l; k++){
		b[k] = b[l-k];
		//fprintf(stdout, "\nb[%3d]= %2.5e", k, b[k]);
		//if( l == 2 )
		//	fprintf(stdout, "\nb[%3d]= %2.5e, %2.5e", k, b[k].d, b[k].e);
	}

	*c = b;

	return;
}

//calculate 2-d combination -- k, k'
void comb2d(bigdouble *** c, int l){
	bigdouble **d;

	if( !(d = (bigdouble **)calloc(l+1, sizeof(bigdouble *))) )
		FPFE1("out of memory for d");

	while(l >= 0){
		comb(d+l, l);
		l--;
	}

	*c = d;
	return;
}

bigdouble bigdouble_mult(bigdouble a, bigdouble b){
	bigdouble c;
	double x;
	int y;

	if( a.d == 0.0 || b.d == 0.0 ){
		c.d = 0.0;
		c.e = 0.0;
	} else {
		c.e = a.e + b.e;
		x = a.d * b.d;

		y = 0;
		if( x >= 10.0 ){
			while(x >= 10.0){
				x /= 10.0;
				y++;
			}
		} else if ( x < 1.0 ) {
			if( x == 0.0 )
				fprintf(stdout, "\nHerem x =0 "), fflush(NULL), exit(0);
			if( x < 0.0 ){
				fprintf(stdout, "\na.d = %f, .e = %e",a.d, a.e);
				fprintf(stdout, "\nb.d = %f, .e = %e",b.d, b.e);
				fprintf(stdout, "\nHerem x < 0 "), fflush(NULL), exit(0);
			}

			while( x < 1.0){
				x *= 10.0;
				y--;
			}
		}

		c.e += y;
		c.d  = x;
	}

	return c;	
}

//calculate p(l, k, k')
void cal_pLKKprime(bigdouble **b, bigdouble ****p, bigdouble ****p1, int l, int n){
	bigdouble ***d;
	bigdouble ***e;
	int i, j, k;
	double x, y;

	if( !(d = (bigdouble ***)calloc(n, sizeof(bigdouble **))) )
		FPFE1("out of memory for d");
	for(i=0; i<n; i++){
		if( !(d[i] = (bigdouble **)calloc(l, sizeof(bigdouble *))) )
			FPFE1("out of memory for d[]");

		for(j=0; j<l; j++)
			if( !(d[i][j] = (bigdouble *)calloc(j+1, sizeof(bigdouble))) )
				FPFE1("out of memory for d[][]");
	}

	if( !(e = (bigdouble ***)calloc(n, sizeof(bigdouble **))) )
		FPFE1("out of memory for e");
	for(i=0; i<n; i++){
		if( !(e[i] = (bigdouble **)calloc(l, sizeof(bigdouble *))) )
			FPFE1("out of memory for e[]");

		for(j=0; j<l; j++)
			if( !(e[i][j] = (bigdouble *)calloc(j+1, sizeof(bigdouble))) )
				FPFE1("out of memory for e[][]");
	}

	//d
	for(i=0; i<l; i++){
	//for(i=1; i<2;i++){
		for(j=0; j<i+1; j++){
			x = (double)(l-i-1);
			//if( x == 0.0 ){
			//	d[0][i][j].d = 0.0;
			//	d[0][i][j].e = 0;
			//} else {
				y=0;
				while(x >= 10.0){
					x /= 10.0;
					y += 1;
				}

				d[0][i][j].d = x * b[i][j].d;
				d[0][i][j].e = y + b[i][j].e;

				if( d[0][i][j].d >= 10.0 ){
					d[0][i][j].d /= 10.0;
					d[0][i][j].e ++;
				}
			//}

			//fprintf(stdout, "\nb[%d][%d].e = %f, .d = %e", i, j, b[i][j].e, b[i][j].d);
			//fprintf(stdout, "\nd[0][%d][%d].e = %f, .d = %e", i, j, d[0][i][j].e, d[0][i][j].d);

			if( d[0][i][j].e >= b[i+1][j+1].e ){				
				y = d[0][i][j].e - b[i+1][j+1].e;
				if( y <= 300 ){				
					d[0][i][j].d += b[i+1][j+1].d / pow(10.0, y);
				}
			} else {
				d[0][i][j].e = b[i+1][j+1].e;
				y = b[i+1][j+1].e - d[0][i][j].e;
				if( y <= 300 ){				
					d[0][i][j].d = b[i+1][j+1].d + d[0][i][j].d / pow(10.0, y);
				}else{
					d[0][i][j].d = b[i+1][j+1].d;
				}
			}

			if( d[0][i][j].d >= 10.0 ){
				d[0][i][j].d /= 10.0;
				d[0][i][j].e  += 1;
			}

			//fprintf(stdout, "\nb[%d][%d].e = %f, .d = %e", i+1, j+1, b[i+1][j+1].e, b[i+1][j+1].d);
			//fprintf(stdout, "\nd[0][%d][%d].e = %f, .d = %e", i, j, d[0][i][j].e, d[0][i][j].d);

			
			x = d[0][i][j].d / b[l][j+1].d;
			y = 0;
			if( x >= 10.0 ){
				while(x >= 10.0){
					x /= 10.0;
					y++;
				}
			} else if ( x < 1.0 ) {
				if( x == 0.0 )
					fprintf(stdout, "\nHere2 x = 0\n"), fflush(NULL), exit(0);
				while( x < 1.0){
					x *= 10.0;
					y--;
				}
			}
			d[0][i][j].e -= b[l][j+1].e;
			d[0][i][j].e += y;	
			d[0][i][j].d  = x;

			//fprintf(stdout, "\nb[%d][%d].e = %f, .d = %e", l, j+1, b[l][j+1].e, b[l][j+1].d);
			//fprintf(stdout, "\nd[0][%d][%d].e = %f, .d = %e", i, j, d[0][i][j].e, d[0][i][j].d);

			//d[0][i][j] = ((l-i-1)*b[i][j] + b[i+1][j+1])/b[l][j+1];
		}
	}

	for(k=1; k<n; k++){
		for(i=0; i<l; i++){
			for(j=0; j<i+1; j++){
				d[k][i][j] = bigdouble_mult(d[k-1][i][j], d[0][i][j]);
			}
		}
	}

	//e
	/*
	for(i=0; i<l; i++){
		for(j=0; j<i+1; j++){
			e[0][i][j] = 1;
		}
	}
	*/
	for(i=0; i<l; i++){
		for(j=0; j<i+1; j++){
			e[1][i][j].e = 0;			
			if( d[0][i][j].e <= -300 ){				
				e[1][i][j].d = 1.0;
			} else {
				e[1][i][j].d = fabs(1.0- pow(10.0, d[0][i][j].e)*d[0][i][j].d);

				if( e[1][i][j].d < 0.0 ){
					fprintf(stdout, "\ni=%d j=%d e.d=%f  b.d = %f, b.e = %e",i,j, e[1][i][j].d, d[0][i][j].d, d[0][i][j].e);
					fprintf(stdout, "\nxerem x < 0 "), fflush(NULL), exit(0);
				}
			}
			//e[1][i][j] = 1-d[0][i][j];
		}
	}

	for(k=2; k<n; k++){
		for(i=0; i<l; i++){
			for(j=0; j<i+1; j++){
				e[k][i][j] = bigdouble_mult(e[k-1][i][j], e[1][i][j]);
			}
		}
	}

	*p = d;
	*p1= e;
	return;
}


//calculate p(n, n', l, k, k')
void cal_pKKN(bigdouble ****e, bigdouble **b, bigdouble ***p, bigdouble ***p1, int l, int n){
	bigdouble ***d;
	int i, j, k;
	double y;

	if( !(d = (bigdouble ***)calloc(l, sizeof(bigdouble **))) )
		FPFE1("out of memory for d");
	for(i=0; i<l; i++){
		if( !(d[i] = (bigdouble **)calloc(i+2, sizeof(bigdouble *))) )
			FPFE1("out of memory for d[]");

		for(j=1; j<i+2; j++)
			if( !(d[i][j] = (bigdouble *)calloc(n+1, sizeof(bigdouble))) )
				FPFE1("out of memory for d[][]");
	}

	for(i=0; i<l; i++){
		for(j=1; j<i+2; j++){			
			d[i][j][n].d = p[n-1][i][j-1].d;
			d[i][j][n].e = p[n-1][i][j-1].e;
		}
	}

	for(i=0; i<l; i++){
		for(k=n-1; k>=1; k--){
			for(j=1; j<i+2; j++){			
				d[i][j][k] = bigdouble_mult(bigdouble_mult(b[n][k],p[k-1][i][j-1]), p1[n-k][i][j-1]);
			}
		}
	}

	for(i=0; i<l; i++){
		for(k=n-1; k>=1; k--){
			for(j=1; j<i+2; j++){
				if( d[i][j][k].e >= d[i][j][k+1].e ){				
					y = d[i][j][k].e - d[i][j][k+1].e;
					if( y <= 300 ){				
						d[i][j][k].d += d[i][j][k+1].d / pow(10.0, y);
					}
				} else {
					d[i][j][k].e = d[i][j][k+1].e;
					y = d[i][j][k+1].e - d[i][j][k].e;
					if( y <= 300 ){										
						d[i][j][k].d = d[i][j][k+1].d + d[i][j][k].d / pow(10.0, y);
					}else{
						d[i][j][k].d = d[i][j][k+1].d;
					}
				}

				if( d[i][j][k].d >= 10.0 ){
					d[i][j][k].d /= 10.0;
					d[i][j][k].e ++;
				}

				//d[i][j][k] += d[i][j][k+1];
			}
		}
	}

	for(i=0; i<l; i++){
		for(j=1; j<i+2; j++){			
			for(k=1; k<=n; k++){
				d[i][j][k] = bigdouble_mult(d[i][j][k], b[l][j]);
			}
		}
	}

	*e = d;
	return;
}


//get evalues in simple format
void cal_EVALUES(bigdouble ***d, int l, int n){
	int i, j, k;

	if( !(allEVALUES = (double ***)calloc(l, sizeof(double **))) )
		FPFE1("out of memory for allEVALUES");
	for(i=0; i<l; i++){
		if( !(allEVALUES[i] = (double **)calloc(i+2, sizeof(double *))) )
			FPFE1("out of memory for allEVALUES[]");

		for(j=1; j<i+2; j++)
			if( !(allEVALUES[i][j] = (double *)calloc(n+1, sizeof(double))) )
				FPFE1("out of memory for allEVALUES[][]");
	}

	for(i=0; i<l; i++){
		for(j=1; j<i+2; j++){			
			for(k=1; k<=n; k++){
				if( d[i][j][k].e >= 1 ){
					allEVALUES[i][j][k] = 10.0;
				} else if ( d[i][j][k].e <= -200 ) {
					allEVALUES[i][j][k] = 0.0;
				} else {
					allEVALUES[i][j][k] = pow(10.0, d[i][j][k].e)*d[i][j][k].d;
				} 
			}
		}
	}

	return;
}

//calculate all e-values
void calcultate_ALL_EVALUE_FAST(int l, int n){
//	FILE *outf;
	int i,j; //, k;
	bigdouble ** comKKprime;
	bigdouble *** pLKKprime; //pLKKprime[0], p(k, k')^1
	bigdouble *** pLKKprime_1;
	bigdouble *** pKKN;  //pKKprimeNprime

//	if( !(outf = fopen("testeval", "w")))
//		FPFE1("Can't open file for writing!");

	//calculate 2-d combination -- k, k'
	comb2d(&comKKprime, l);

	//calculate p(l, k, k') and powers
	cal_pLKKprime(comKKprime, &pLKKprime, &pLKKprime_1, l, n);

	//calculate e values
	cal_pKKN(&pKKN, comKKprime, pLKKprime, pLKKprime_1, l, n);

	//get evalues in simple format
	cal_EVALUES(pKKN, l, n);

	/*
	for(i=0; i<l; i++){
		fprintf(outf, "\n\n>l= %d", i+1);
		for(j=1; j<=i+1; j++){
			fprintf(outf, "\n%4d", j);
			for(k=1; k<=n; k++){
				fprintf(outf, " %2.5f e%4d", pKKN[i][j][k].d, (int)pKKN[i][j][k].e);
			}
		}
	}
	fclose(outf);
	*/

	//Free space
	if( comKKprime ){
		for(i=0; i<=l; i++)
			if( comKKprime[i] )
				free(comKKprime[i]);
		free(comKKprime);
	}

	if( pLKKprime ){
		for(i=0; i<n; i++){
			for(j=0; j<l; j++){
				if( pLKKprime[i][j] )
					free( pLKKprime[i][j] );
			}
			free( pLKKprime[i] );
		}
		free( pLKKprime );
	}

	if( pLKKprime_1 ){
		for(i=0; i<n; i++){
			for(j=0; j<l; j++){
				if( pLKKprime_1[i][j] )
					free( pLKKprime_1[i][j] );
			}
			free( pLKKprime_1[i] );
		}
		free( pLKKprime_1 );
	}

	if( pKKN ){
		for(i=0; i<l; i++){
			for(j=1; j<i+2; j++)
				if( pKKN[i][j] )
					free( pKKN[i][j] );
			free( pKKN[i] );
		}
		free( pKKN );
	}

	return;
}


//calculate p(l, k, k')   --- for circular chromosomes
void comput_pLKKprime_circular(bigdouble **b, bigdouble ****p, bigdouble ****p1, int l, int n, int r){
	bigdouble ***d;
	bigdouble ***e;
	int i, j, k;
	double x, y;

	if( !(d = (bigdouble ***)calloc(n, sizeof(bigdouble **))) )
		FPFE1("out of memory for d");
	for(i=0; i<n; i++){
		if( !(d[i] = (bigdouble **)calloc(r, sizeof(bigdouble *))) )
			FPFE1("out of memory for d[]");

		for(j=0; j<r; j++)
			if( !(d[i][j] = (bigdouble *)calloc(j+1, sizeof(bigdouble))) )
				FPFE1("out of memory for d[][]");
	}

	if( !(e = (bigdouble ***)calloc(n, sizeof(bigdouble **))) )
		FPFE1("out of memory for e");
	for(i=0; i<n; i++){
		if( !(e[i] = (bigdouble **)calloc(r, sizeof(bigdouble *))) )
			FPFE1("out of memory for e[]");

		for(j=0; j<r; j++)
			if( !(e[i][j] = (bigdouble *)calloc(j+1, sizeof(bigdouble))) )
				FPFE1("out of memory for e[][]");
	}

	//d
	for(i=0; i<r; i++){
		for(j=0; j<i+1; j++){
			//x = (double)(l-i-1);
			x = (double)l;
			y=0;
			while(x >= 10.0){
				x /= 10.0;
				y += 1;
			}

			d[0][i][j].d = x * b[i][j].d;
			d[0][i][j].e = y + b[i][j].e;

			if( d[0][i][j].d >= 10.0 ){
				d[0][i][j].d /= 10.0;
				d[0][i][j].e ++;
			}

			/*
			if( d[0][i][j].e >= b[i+1][j+1].e ){				
				y = d[0][i][j].e - b[i+1][j+1].e;
				if( y <= 300 ){				
					d[0][i][j].d += b[i+1][j+1].d / pow(10.0, y);
				}
			} else {
				d[0][i][j].e = b[i+1][j+1].e;
				y = b[i+1][j+1].e - d[0][i][j].e;
				if( y <= 300 ){				
					d[0][i][j].d = b[i+1][j+1].d + d[0][i][j].d / pow(10.0, y);
				}else{
					d[0][i][j].d = b[i+1][j+1].d;
				}
			}

			if( d[0][i][j].d >= 10.0 ){
				d[0][i][j].d /= 10.0;
				d[0][i][j].e  += 1;
			}
			*/
			
			x = d[0][i][j].d / b[l][j+1].d;
			y = 0;
			if( x >= 10.0 ){
				while(x >= 10.0){
					x /= 10.0;
					y++;
				}
			} else if ( x < 1.0 ) {
				if( x == 0.0 )
					fprintf(stdout, "\nHere2 x = 0\n"), fflush(NULL), exit(0);
				while( x < 1.0){
					x *= 10.0;
					y--;
				}
			}
			d[0][i][j].e -= b[l][j+1].e;
			d[0][i][j].e += y;	
			d[0][i][j].d  = x;
		}
	}

	for(k=1; k<n; k++){
		for(i=0; i<r; i++){
			for(j=0; j<i+1; j++){
				d[k][i][j] = bigdouble_mult(d[k-1][i][j], d[0][i][j]);
			}
		}
	}

	//e
	for(i=0; i<r; i++){
		for(j=0; j<i+1; j++){
			e[1][i][j].e = 0;			
			if( d[0][i][j].e <= -300 ){				
				e[1][i][j].d = 1.0;
			} else {
				e[1][i][j].d = fabs(1.0- pow(10.0, d[0][i][j].e)*d[0][i][j].d);

				if( e[1][i][j].d < 0.0 ){
					fprintf(stdout, "\ni=%d j=%d e.d=%f  b.d = %f, b.e = %e",i,j, e[1][i][j].d, d[0][i][j].d, d[0][i][j].e);
					fprintf(stdout, "\nxerem x < 0 "), fflush(NULL), exit(0);
				}
			}
		}
	}

	for(k=2; k<n; k++){
		for(i=0; i<r; i++){
			for(j=0; j<i+1; j++){
				e[k][i][j] = bigdouble_mult(e[k-1][i][j], e[1][i][j]);
			}
		}
	}

	*p = d;
	*p1= e;
	return;
}


//calculate p(l, k, k')   --- for linear chromosomes
void comput_pLKKprime(bigdouble **b, bigdouble ****p, bigdouble ****p1, int l, int n, int r){
	bigdouble ***d;
	bigdouble ***e;
	int i, j, k;
	double x, y;

	if( !(d = (bigdouble ***)calloc(n, sizeof(bigdouble **))) )
		FPFE1("out of memory for d");
	for(i=0; i<n; i++){
		if( !(d[i] = (bigdouble **)calloc(r, sizeof(bigdouble *))) )
			FPFE1("out of memory for d[]");

		for(j=0; j<r; j++)
			if( !(d[i][j] = (bigdouble *)calloc(j+1, sizeof(bigdouble))) )
				FPFE1("out of memory for d[][]");
	}

	if( !(e = (bigdouble ***)calloc(n, sizeof(bigdouble **))) )
		FPFE1("out of memory for e");
	for(i=0; i<n; i++){
		if( !(e[i] = (bigdouble **)calloc(r, sizeof(bigdouble *))) )
			FPFE1("out of memory for e[]");

		for(j=0; j<r; j++)
			if( !(e[i][j] = (bigdouble *)calloc(j+1, sizeof(bigdouble))) )
				FPFE1("out of memory for e[][]");
	}

	//d
	for(i=0; i<r; i++){
		for(j=0; j<i+1; j++){
			x = (double)(l-i-1);
			y=0;
			while(x >= 10.0){
				x /= 10.0;
				y += 1;
			}

			d[0][i][j].d = x * b[i][j].d;
			d[0][i][j].e = y + b[i][j].e;

			if( d[0][i][j].d >= 10.0 ){
				d[0][i][j].d /= 10.0;
				d[0][i][j].e ++;
			}

			if( d[0][i][j].e >= b[i+1][j+1].e ){				
				y = d[0][i][j].e - b[i+1][j+1].e;
				if( y <= 300 ){				
					d[0][i][j].d += b[i+1][j+1].d / pow(10.0, y);
				}
			} else {
				d[0][i][j].e = b[i+1][j+1].e;
				y = b[i+1][j+1].e - d[0][i][j].e;
				if( y <= 300 ){				
					d[0][i][j].d = b[i+1][j+1].d + d[0][i][j].d / pow(10.0, y);
				}else{
					d[0][i][j].d = b[i+1][j+1].d;
				}
			}

			if( d[0][i][j].d >= 10.0 ){
				d[0][i][j].d /= 10.0;
				d[0][i][j].e  += 1;
			}
			
			x = d[0][i][j].d / b[l][j+1].d;
			y = 0;
			if( x >= 10.0 ){
				while(x >= 10.0){
					x /= 10.0;
					y++;
				}
			} else if ( x < 1.0 ) {
				if( x == 0.0 )
					fprintf(stdout, "\nHere2 x = 0\n"), fflush(NULL), exit(0);
				while( x < 1.0){
					x *= 10.0;
					y--;
				}
			}
			d[0][i][j].e -= b[l][j+1].e;
			d[0][i][j].e += y;	
			d[0][i][j].d  = x;
		}
	}

	for(k=1; k<n; k++){
		for(i=0; i<r; i++){
			for(j=0; j<i+1; j++){
				d[k][i][j] = bigdouble_mult(d[k-1][i][j], d[0][i][j]);
			}
		}
	}

	//e
	for(i=0; i<r; i++){
		for(j=0; j<i+1; j++){
			e[1][i][j].e = 0;			
			if( d[0][i][j].e <= -300 ){				
				e[1][i][j].d = 1.0;
			} else {
				e[1][i][j].d = fabs(1.0- pow(10.0, d[0][i][j].e)*d[0][i][j].d);

				if( e[1][i][j].d < 0.0 ){
					fprintf(stdout, "\ni=%d j=%d e.d=%f  b.d = %f, b.e = %e",i,j, e[1][i][j].d, d[0][i][j].d, d[0][i][j].e);
					fprintf(stdout, "\nxerem x < 0 "), fflush(NULL), exit(0);
				}
			}
		}
	}

	for(k=2; k<n; k++){
		for(i=0; i<r; i++){
			for(j=0; j<i+1; j++){
				e[k][i][j] = bigdouble_mult(e[k-1][i][j], e[1][i][j]);
			}
		}
	}

	*p = d;
	*p1= e;
	return;
}


//calculate p(n, n', l, k, k')
void comput_pKKN(bigdouble ****e, bigdouble **b, bigdouble ***p, bigdouble ***p1, int l, int n, int r){
	bigdouble ***d;
	int i, j, k;
	double y;

	if( !(d = (bigdouble ***)calloc(r, sizeof(bigdouble **))) )
		FPFE1("out of memory for d");
	for(i=0; i<r; i++){
		if( !(d[i] = (bigdouble **)calloc(i+2, sizeof(bigdouble *))) )
			FPFE1("out of memory for d[]");

		for(j=1; j<i+2; j++)
			if( !(d[i][j] = (bigdouble *)calloc(n+1, sizeof(bigdouble))) )
				FPFE1("out of memory for d[][]");
	}

	for(i=0; i<r; i++){
		for(j=1; j<i+2; j++){			
			d[i][j][n].d = p[n-1][i][j-1].d;
			d[i][j][n].e = p[n-1][i][j-1].e;
		}
	}

	for(i=0; i<r; i++){
		for(k=n-1; k>=1; k--){
			for(j=1; j<i+2; j++){			
				d[i][j][k] = bigdouble_mult(bigdouble_mult(b[n][k],p[k-1][i][j-1]), p1[n-k][i][j-1]);
			}
		}
	}

	for(i=0; i<r; i++){
		for(k=n-1; k>=1; k--){
			for(j=1; j<i+2; j++){
				if( d[i][j][k].e >= d[i][j][k+1].e ){				
					y = d[i][j][k].e - d[i][j][k+1].e;
					if( y <= 300 ){				
						d[i][j][k].d += d[i][j][k+1].d / pow(10.0, y);
					}
				} else {
					d[i][j][k].e = d[i][j][k+1].e;
					y = d[i][j][k+1].e - d[i][j][k].e;
					if( y <= 300 ){										
						d[i][j][k].d = d[i][j][k+1].d + d[i][j][k].d / pow(10.0, y);
					}else{
						d[i][j][k].d = d[i][j][k+1].d;
					}
				}

				if( d[i][j][k].d >= 10.0 ){
					d[i][j][k].d /= 10.0;
					d[i][j][k].e ++;
				}

				//d[i][j][k] += d[i][j][k+1];
			}
		}
	}

	for(i=0; i<r; i++){
		for(j=1; j<i+2; j++){			
			for(k=1; k<=n; k++){
				d[i][j][k] = bigdouble_mult(d[i][j][k], b[l][j]);
			}
		}
	}

	*e = d;
	return;
}


//get evalues in simple format
void comput_EVALUES(bigdouble ***d, int l, int n, int r){
	int i, j, k;

	if( !(allEVALUES = (double ***)calloc(r, sizeof(double **))) )
		FPFE1("out of memory for allEVALUES");
	for(i=0; i<r; i++){
		if( !(allEVALUES[i] = (double **)calloc(i+2, sizeof(double *))) )
			FPFE1("out of memory for allEVALUES[]");

		for(j=1; j<i+2; j++)
			if( !(allEVALUES[i][j] = (double *)calloc(n+1, sizeof(double))) )
				FPFE1("out of memory for allEVALUES[][]");
	}

	for(i=0; i<r; i++){
		for(j=1; j<i+2; j++){			
			for(k=1; k<=n; k++){
				if( d[i][j][k].e >= 10 ){
					allEVALUES[i][j][k] = 1e10;
				} else if ( d[i][j][k].e <= -300 ) {
					allEVALUES[i][j][k] = 0.0;
				} else {
					allEVALUES[i][j][k] = pow(10.0, d[i][j][k].e)*d[i][j][k].d;
				} 
			}
		}
	}

	return;
}


//calculate all e-values
void comput_ALL_EVALUE(int l, int n, int r){
	int i,j;
	bigdouble ** comKKprime;
	bigdouble *** pLKKprime; //pLKKprime[0], p(k, k')^1
	bigdouble *** pLKKprime_1;
	bigdouble *** pKKN;  //pKKprimeNprime

	//calculate 2-d combination -- k, k'
	comb2d(&comKKprime, l);

	if ( IF_CIRCULAR_GENOME == 0 ){
		//calculate p(l, k, k') and powers  --- for linear chromosomes
		comput_pLKKprime(comKKprime, &pLKKprime, &pLKKprime_1, l, n, r);
	} else {
		//calculate p(l, k, k')   --- for circular chromosomes
		comput_pLKKprime_circular(comKKprime, &pLKKprime, &pLKKprime_1, l, n, r);
	}

	//calculate e values
	comput_pKKN(&pKKN, comKKprime, pLKKprime, pLKKprime_1, l, n, r);

	//get evalues in simple format
	comput_EVALUES(pKKN, l, n, r);


	if( comKKprime ){
		for(i=0; i<=l; i++)
			if( comKKprime[i] )
				free(comKKprime[i]);
		free(comKKprime);
	}

	if( pLKKprime ){
		for(i=0; i<n; i++){
			for(j=0; j<r; j++){
				if( pLKKprime[i][j] )
					free( pLKKprime[i][j] );
			}
			free( pLKKprime[i] );
		}
		free( pLKKprime );
	}

	if( pLKKprime_1 ){
		for(i=0; i<n; i++){
			for(j=0; j<r; j++){
				if( pLKKprime_1[i][j] )
					free( pLKKprime_1[i][j] );
			}
			free( pLKKprime_1[i] );
		}
		free( pLKKprime_1 );
	}

	if( pKKN ){
		for(i=0; i<r; i++){
			for(j=1; j<i+2; j++)
				if( pKKN[i][j] )
					free( pKKN[i][j] );
			free( pKKN[i] );
		}
		free( pKKN );
	}

	return;
}



//reorganize the output results
void reorganize_output(int ng, char *tfname, char * ofname){
	FILE *inf, *outf;
	char  buff[4096];
	int   n, i, y2;

	rnode *head, *tail, *curr, *burr, *durr, *turr, *newnode;

	if( !(head = malloc(sizeof(rnode))) )
		FPFE1("out of memory for head\n");
	head->next = NULL;
	n = ng + 2;

	if( !(inf = fopen(tfname, "r")) )
		return;
		//FPFE2("File %s can't be open for reading syn\n", tfname);
	if( !(outf = fopen(ofname, "a")) )
		FPFE2("File %s can't be open for writing\n", ofname);

	//read the data
	 while( fgets(buff, 4096, inf) ){
		 if( strstr(buff, "Cluster") ){
			 newnode = malloc(sizeof(rnode));
			 newnode->next = NULL;

			 //read in all data for the new node
			 if( !(newnode->s = malloc(n*sizeof(char *))) )
				 FPFE1("out of memory for newnode->s\n");
			 if( !((newnode->s)[0] = malloc((strlen(buff)+1)*sizeof(char))) )
				 FPFE1("out of memory for newnode->s[0]\n");

			 strcpy((newnode->s)[0], buff);

			 for(i=1; i<n; i++){
				 if( !fgets(buff, 4096, inf) )
					 fprintf(stdout, "data error after the line %s", (newnode->s)[i-1]), fflush(NULL), exit(0);
					 //FPFE2("data error after the line %s", (newnode->s)[i-1]);
				 if( strstr(buff, "@@") ){
					 (newnode->s)[i] = NULL;
				 } else {
					if( !((newnode->s)[i] = malloc((strlen(buff)+1)*sizeof(char))) )
						 FPFE1("out of memory for newnode->s[]\n");
					strcpy((newnode->s)[i], buff);
				 }
			 }

			 //determine eval, l, n
			 if( sscanf((newnode->s)[1], " Expect = %le, Size = %d/%*s Appear = %d/%*s", &(newnode->eval), &(newnode->l), &(newnode->n)) !=3){
				 FPFE2("data error in line:\n %s", (newnode->s)[1]);
			 }
			 //determine set
			 if( ( (newnode->set) = malloc(newnode->l * sizeof(int))) ){
				 int xh;//, xi, xj, xp;
				 char *p;

				 //read in the array
				 strcpy(buff, (newnode->s)[0]);
				 if( !(p = strstr(buff, ":")) )
					 FPFE2("data error (no :) in line:\n %s", buff);
				 p++;
				 p = strtok(p, " ,");
				 (newnode->set)[0] = atoi(p);
				 xh =1;
				 while( (p = strtok(NULL, " ,")) ){
					(newnode->set)[xh++] = atoi(p);
				 }
				 if( xh != newnode->l )
					 FPFE2("xh != newnode->l in cluster %s\n", (newnode->s)[0]);

				 /*//sort the array
				 for(xi=0; xi<newnode->l; xi++){
					 xh = (newnode->set)[xi];
					 for(xj=xi+1; xj<newnode->l; xj++){
						 if( (newnode->set)[xj] == xh )
							 FPFE2("(newnode->set)[xj] == xh  in cluster %s\n", (newnode->s)[0]);
						 if( (newnode->set)[xj] < xh ){
							 xh = (newnode->set)[xj];
							 xp = xj;
						 }
					 }
					 if( xh < (newnode->set)[xi] ){
						(newnode->set)[xp] = (newnode->set)[xi];
						(newnode->set)[xi] = xh;
					 }
				 }
				 */
			 } else
				 FPFE1("out of memory for newnode->set\n");

			 //determine  chr low high
			 if( !( (newnode->chr) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->chr\n");
			 if( !( (newnode->low) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->low\n");
			 if( !( (newnode->high) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->high\n");

			 for(i=2; i<n; i++){
				 if( (newnode->s)[i] == NULL ){
					 (newnode->chr)[i-2] = -1;
					 (newnode->low)[i-2] = -1;
					 (newnode->high)[i-2] = -1;
				 } else {
					 if( sscanf((newnode->s)[i], " %*s (Chr:%d S%d E%d): %*s", &((newnode->chr)[i-2]), &((newnode->low)[i-2]), &((newnode->high)[i-2])) != 3 )
						 FPFE2("data error in line:\n%s", (newnode->s)[i]);
				 }
			 }

			 //add the new node to the linked list
			 burr = head;
			 curr = head->next;

			 while( curr != NULL && curr->eval <= newnode->eval ){
				 burr = burr->next;
				 curr = curr->next;
			 }

			 if( curr == NULL ){
				 burr->next = newnode;
				 tail = newnode;
			 } else {
				 burr->next = newnode;
				 newnode->next = curr;
			 }
		 }
	 }

	 //remove repeats and subsets
	 if( head->next != NULL && (head->next)->next != NULL ){
		 int yc, yd, y1;
		 burr = head->next;
		 curr = (head->next)->next;
		 while( curr != NULL ){
			 durr = head->next;
			 while( durr != curr ){
				 if( durr->l >= curr->l ){
					 yc = yd = 0;
					 while( yc < curr->l && yd < durr->l ){
						 if( curr->set[yc] == durr->set[yd] ){
							 yc++;
							 yd++;
						 } else if ( curr->set[yc] < durr->set[yd] ){
							 break;
						 } else {
							 yd++;
						 }
					 }

					 if( yc == curr->l ){ // possible -- subset or same
						for( y1=0; y1<ng; y1++ ){
							if( curr->chr[y1] != -1 ){
								if( curr->chr[y1] == durr->chr[y1] && curr->low[y1]>=durr->low[y1] && curr->high[y1]<=durr->high[y1] ){
								} else
									break;
							}
						}
						if( y1 == ng ){ // subset or same
							turr = curr;
							burr->next = curr->next;
							curr = curr->next;

							//free space
							if( turr->set )
								free(turr->set);
							if( turr->chr )
								free(turr->chr);
							if( turr->low )
								free(turr->low);
							if( turr->high )
								free(turr->high);

							if( turr->s ){
								for( y2=0; y2<n; y2++ )
									if( turr->s[y2] )
										free( turr->s[y2] );
								free(turr->s);
							}
							if( turr )
								free(turr);

							break;
						}

					 }
				 }

				 durr = durr->next;
			 }

			 if ( durr == curr ) {
				 burr=burr->next;
				 curr=curr->next;
			 }
		 }
	 }

	//output new results
	curr = head->next;
	while(curr){
		 turr = curr;
		 for(i=0; i<n; i++)
			if( curr->s[i] ){
				fprintf(outf, "%s", curr->s[i]);
			} else {
				fprintf(outf, "@@\n");
			}
		 fprintf(outf, "\n\n");
		 curr = curr->next;

		 //free space
		if( turr->set )
			free(turr->set);
		if( turr->chr )
			free(turr->chr);
		if( turr->low )
			free(turr->low);
		if( turr->high )
			free(turr->high);

		for( y2=0; y2<n; y2++ )
			if( turr->s[y2] )
				free( turr->s[y2] );
		if( turr->s )
			free(turr->s);
		if( turr )
			free(turr);
	}

	fclose(inf);
	fclose(outf);

	return;
}


//reorganize the output results
void reorganize_output_2(int ng, char *tfname){
	FILE *inf, *outf;
	char  buff[4096];
	int   n, i, y2;

	rnode *head, *tail, *curr, *burr, *durr, *turr, *newnode;

	if( !(head = malloc(sizeof(rnode))) )
		FPFE1("out of memory for head\n");
	head->next = NULL;
	n = ng + 2;

	if( !(inf = fopen(tfname, "r")) )
		return;

	//read the data
	 while( fgets(buff, 4096, inf) ){
		 if( strstr(buff, "Cluster") ){
			 newnode = malloc(sizeof(rnode));
			 newnode->next = NULL;

			 //read in all data for the new node
			 if( !(newnode->s = malloc(n*sizeof(char *))) )
				 FPFE1("out of memory for newnode->s\n");
			 if( !((newnode->s)[0] = malloc((strlen(buff)+1)*sizeof(char))) )
				 FPFE1("out of memory for newnode->s[0]\n");

			 strcpy((newnode->s)[0], buff);

			 for(i=1; i<n; i++){
				 if( !fgets(buff, 4096, inf) )
					 fprintf(stdout, "data error after the line %s", (newnode->s)[i-1]), fflush(NULL), exit(0);
					 //FPFE2("data error after the line %s", (newnode->s)[i-1]);
				 if( strstr(buff, "@@") ){
					 (newnode->s)[i] = NULL;
				 } else {
					if( !((newnode->s)[i] = malloc((strlen(buff)+1)*sizeof(char))) )
						 FPFE1("out of memory for newnode->s[]\n");
					strcpy((newnode->s)[i], buff);
				 }
			 }

			 //determine eval, l, n
			 if( sscanf((newnode->s)[1], " Expect = %le, Size = %d/%*s Appear = %d/%*s", &(newnode->eval), &(newnode->l), &(newnode->n)) !=3){
				 FPFE2("data error in line:\n %s", (newnode->s)[1]);
			 }
			 //determine set
			 if( ( (newnode->set) = malloc(newnode->l * sizeof(int))) ){
				 int xh;//, xi, xj, xp;
				 char *p;

				 //read in the array
				 strcpy(buff, (newnode->s)[0]);
				 if( !(p = strstr(buff, ":")) )
					 FPFE2("data error (no :) in line:\n %s", buff);
				 p++;
				 p = strtok(p, " ,");
				 (newnode->set)[0] = atoi(p);
				 xh =1;
				 while( (p = strtok(NULL, " ,")) ){
					(newnode->set)[xh++] = atoi(p);
				 }
				 if( xh != newnode->l )
					 FPFE2("xh != newnode->l in cluster %s\n", (newnode->s)[0]);

				 /*//sort the array
				 for(xi=0; xi<newnode->l; xi++){
					 xh = (newnode->set)[xi];
					 for(xj=xi+1; xj<newnode->l; xj++){
						 if( (newnode->set)[xj] == xh )
							 FPFE2("(newnode->set)[xj] == xh  in cluster %s\n", (newnode->s)[0]);
						 if( (newnode->set)[xj] < xh ){
							 xh = (newnode->set)[xj];
							 xp = xj;
						 }
					 }
					 if( xh < (newnode->set)[xi] ){
						(newnode->set)[xp] = (newnode->set)[xi];
						(newnode->set)[xi] = xh;
					 }
				 }
				 */
			 } else
				 FPFE1("out of memory for newnode->set\n");

			 //determine  chr low high
			 if( !( (newnode->chr) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->chr\n");
			 if( !( (newnode->low) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->low\n");
			 if( !( (newnode->high) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->high\n");

			 for(i=2; i<n; i++){
				 if( (newnode->s)[i] == NULL ){
					 (newnode->chr)[i-2] = -1;
					 (newnode->low)[i-2] = -1;
					 (newnode->high)[i-2] = -1;
				 } else {
					 if( sscanf((newnode->s)[i], " %*s (Chr:%d S%d E%d): %*s", &((newnode->chr)[i-2]), &((newnode->low)[i-2]), &((newnode->high)[i-2])) != 3 )
						 FPFE2("data error in line:\n%s", (newnode->s)[i]);
				 }
			 }

			 //add the new node to the linked list
			 burr = head;
			 curr = head->next;

			 while( curr != NULL && curr->eval <= newnode->eval ){
				 burr = burr->next;
				 curr = curr->next;
			 }

			 if( curr == NULL ){
				 burr->next = newnode;
				 tail = newnode;
			 } else {
				 burr->next = newnode;
				 newnode->next = curr;
			 }
		 }
	 }

	 //remove repeats and subsets
	 if( head->next != NULL && (head->next)->next != NULL ){
		 int yc, yd, y1;
		 burr = head->next;
		 curr = (head->next)->next;
		 while( curr != NULL ){
			 durr = head->next;
			 while( durr != curr ){
				 if( durr->l >= curr->l ){
					 yc = yd = 0;
					 while( yc < curr->l && yd < durr->l ){
						 if( curr->set[yc] == durr->set[yd] ){
							 yc++;
							 yd++;
						 } else if ( curr->set[yc] < durr->set[yd] ){
							 break;
						 } else {
							 yd++;
						 }
					 }

					 if( yc == curr->l ){ // possible -- subset or same
						for( y1=0; y1<ng; y1++ ){
							if( curr->chr[y1] != -1 ){
								if( curr->chr[y1] == durr->chr[y1] && curr->low[y1]>=durr->low[y1] && curr->high[y1]<=durr->high[y1] ){
								} else
									break;
							}
						}
						if( y1 == ng ){ // subset or same
							turr = curr;
							burr->next = curr->next;
							curr = curr->next;

							//free space
							if( turr->set )
								free(turr->set);
							if( turr->chr )
								free(turr->chr);
							if( turr->low )
								free(turr->low);
							if( turr->high )
								free(turr->high);

							for( y2=0; y2<n; y2++ )
								if( turr->s[y2] )
									free( turr->s[y2] );
							if( turr->s )
								free(turr->s);
							if( turr )
								free(turr);

							break;
						}

					 }
				 }

				 durr = durr->next;
			 }

			 if ( durr == curr ) {
				 burr=burr->next;
				 curr=curr->next;
			 }
		 }
	 }

	 //remove old version
	 fclose(inf);
	 remove(tfname);

	 if( !(outf = fopen(tfname, "w")) )
		return;

	 //output new results
	 curr = head->next;
	 while(curr){
		 turr = curr;
		 for(i=0; i<n; i++)
			if( curr->s[i] ){
				fprintf(outf, "%s", curr->s[i]);
			} else {
				fprintf(outf, "@@\n");
			}
		 fprintf(outf, "\n\n");
		 curr = curr->next;

		 //free space
		if( turr->set )
			free(turr->set);
		if( turr->chr )
			free(turr->chr);
		if( turr->low )
			free(turr->low);
		if( turr->high )
			free(turr->high);

		for( y2=0; y2<n; y2++ )
			if( turr->s[y2] )
				free( turr->s[y2] );
		if( turr->s )
			free(turr->s);
		if( turr )
			free(turr);
	 }

	fclose(outf);

	return;
}







